# -*- Mode: Python; coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2010 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s):   George Kussumoto    <george@async.com.br>
##

import gtk
import gobject

import os

from kiwi.log import Logger
from kiwi.ui.dialogs import HIGAlertDialog

from stoqlib.database.runtime import (get_connection, new_transaction,
                                     finish_transaction)
from stoqlib.domain.events import CreatePaymentEvent, CheckECFStateEvent
from stoqlib.enums import CreatePaymentStatus
from stoqlib.gui.base.dialogs import run_dialog
from stoqlib.gui.events import StartApplicationEvent
from stoqlib.lib.message import info, warning
from stoqlib.lib.osutils import get_application_dir
from stoqlib.lib.parameters import sysparam

from tefdomain import TEFClient
from tefclientdialog import TEFClientDialog
from vespague import Vespague
from teftranslation import tef_gettext as _

log = Logger("stoqlib.plugins.tef.tefui")


class TEFUI(object):
    def __init__(self):
        self._tef_initialized = False
        self.conn = get_connection()
        self._pending_flag = os.path.join(get_application_dir(),
                                          'tef_pending_flag')

        StartApplicationEvent.connect(self._on_StartApplicationEvent)
        CreatePaymentEvent.connect(self._on_CreatePaymentEvent)

        self._vespague = Vespague()

    #
    # Private
    #

    def _check_pending_flag(self):
        log.debug('Checking pending tef transactions flag')
        tag_exists = os.access(self._pending_flag, os.F_OK)
        while tag_exists:
            warning('Há pelo menos uma transação PENDENTE. Favor realizar o '
                    'DESFAZIMENTO. Cancelar o cupom fiscal !')

            self._run_pending_transactions_dialog()
            tag_exists = os.access(self._pending_flag, os.F_OK)

    def _add_ui_menus(self, appname, uimanager):
        if appname == 'pos':
            self._initialize_tef_client()
            self._add_pos_menus(uimanager)
        elif appname == 'admin':
            self._add_admin_menus(uimanager)

    def _add_pos_menus(self, uimanager):
        ui_string = """<ui>
          <menubar name="menubar">
            <placeholder name="ExtraMenu">
              <menu action="TEFMenu">
                <menuitem action="ReprintSale" name="ReprintSale"/>
                <menuitem action="CancelSale" name="CancelSale"/>
                <menuitem action="PedingTransactions" name="PedingTransactions"/>
              </menu>
            </placeholder>
          </menubar>
        </ui>"""

        ag = gtk.ActionGroup('ECFMenuActions')
        ag.add_actions([
            ('TEFMenu', None, 'TEF'),
            ('ReprintSale', None, _('Reprint Sale'),
             None, None, self._on_ReprintSale__activate),
            ('CancelSale', None, _('Cancel Sale'), None, None,
             self._on_CancelSale__activate),
            ('PedingTransactions', None, _(u'Pending Transactions'), None,
             None, self._on_PedingTransactions__activate),
            ])

        uimanager.insert_action_group(ag, 0)
        uimanager.add_ui_from_string(ui_string)

    def _add_admin_menus(self, uimanager):
        ui_string = """<ui>
          <menubar name="menubar">
            <menu action="ConfigureMenu">
            <placeholder name="PluginSettings">
              <menuitem action="ConfigureTEFClient"/>
            </placeholder>
            </menu>
          </menubar>
        </ui>"""

        ag = gtk.ActionGroup('TEFMenuActions')
        ag.add_actions([
            ('TEFMenu', None, 'TEF'),
            ('ConfigureTEFClient', None, _(u'Configure TEF client...'), None,
             None, self._on_ConfigureTEFClient__activate),])
        uimanager.insert_action_group(ag, 0)
        uimanager.add_ui_from_string(ui_string)

    def _initialize_tef_client(self):
        tefclients = TEFClient.select(connection=self.conn).limit(1)
        if tefclients:
            log.debug('Initializing tef')
            #FIXME: handle errors
            self.process = tefclients[0].run()
            if self.process and not sysparam(self.conn).DEMO_MODE:
                self._display_startup_dialog()
                gobject.io_add_watch(self.process.stdout.fileno(), gobject.IO_IN,
                                     self._read_vespague_client_output)
            else:
                self._tef_initialized = True
        else:
            info(_(u'Check the TEF client settings, it failed to start.'))

    def _read_vespague_client_output(self, source, condition):
        line = self.process.stdout.readline()[:-1]
        if line:
            log.debug(line)

        check1 = 'Aplicação já em execução'
        check2 = 'TERMINAL ATIVADO'
        ready = check1 in line or check2 in line
        if ready and not self._tef_initialized:
            log.debug('TEF client is ready')
            self._tef_initialized = True
            self._hide_startup_dialog()
            # We should always check for pending transactions when starting
            # the app
            self._check_pending_flag()
            CheckECFStateEvent.emit()

        return True

    def _display_startup_dialog(self):
        d = HIGAlertDialog(parent=None, flags=gtk.DIALOG_MODAL,
                           type=gtk.MESSAGE_INFO,
                           buttons=gtk.BUTTONS_NONE)
        d.set_primary(_('Waiting TEF to start'))
        # Prevents ESC to close the dialog
        d.connect("delete-event", lambda x,y: True)
        toplevel = d.get_toplevel()
        toplevel.set_deletable(False)
        d.show()
        self._starting_tef_dialog = d

    def _hide_startup_dialog(self):
        self._starting_tef_dialog.hide()
        self._starting_tef_dialog.destroy()

    def _run_sale_card_dialog(self, sale, payment_method, value=None):
        log.debug('Running sale card dialog for sale %s' % sale)
        if not self._tef_initialized:
            info(_('TEF not initialized. Check the TEF client settings.'))
            return False

        retval = self._vespague.run_sale_card_dialog(sale, payment_method,
                                                     value)
        return retval

    def _run_cancel_sale_dialog(self):
        log.debug('Running cancel sale dialog')
        if not self._tef_initialized:
            info(_('TEF not initialized. Check the TEF client settings.'))
            return False

        retval = self._vespague.run_cancel_sale_dialog()
        return retval

    def _run_reprint_sale_dialog(self):
        log.debug('Running reprint sale dialog')
        if not self._tef_initialized:
            info(_('TEF not initialized. Check the TEF client settings.'))
            return False

        retval = self._vespague.run_reprint_sale_dialog()
        return retval

    def _run_pending_transactions_dialog(self):
        log.debug('Running pending transations dialog')
        if not self._tef_initialized:
            info(_('TEF not initialized. Check the TEF client settings.'))
            return False

        retval = self._vespague.run_pending_transactions_dialog()
        return retval

    #
    # Events
    #

    def _on_StartApplicationEvent(self, appname, app):
        self._add_ui_menus(appname, app.main_window.uimanager)

    def _on_CreatePaymentEvent(self, payment_method, sale, conn, value=None):
        if payment_method.method_name == 'card':
            retval = self._run_sale_card_dialog(sale, payment_method, value)
            if retval:
                return CreatePaymentStatus.SUCCESS
            else:
                return CreatePaymentStatus.FAIL
        # Fallback.
        return CreatePaymentStatus.UNHANDLED

    #
    # Callbacks
    #

    def _on_ConfigureTEFClient__activate(self, action):
        trans = new_transaction()
        retval = run_dialog(TEFClientDialog, None, trans)
        finish_transaction(trans, retval)
        trans.close()

    def _on_ReprintSale__activate(self, action):
        self._run_reprint_sale_dialog()

    def _on_CancelSale__activate(self, action):
        self._run_cancel_sale_dialog()

    def _on_PedingTransactions__activate(self, action):
        self._run_pending_transactions_dialog()
