# -*- coding: utf-8 -*-
# vi:si:et:sw=4:sts=4:ts=4

##
## Copyright (C) 2010 Async Open Source <http://www.async.com.br>
## All rights reserved
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU Lesser General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU Lesser General Public License for more details.
##
## You should have received a copy of the GNU Lesser General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., or visit: http://www.gnu.org/.
##
## Author(s): George Kussumoto      <george@async.com.br>
##            Ronaldo Maia          <romaia@async.com.br>
##

import datetime
import dateutil.parser
import re
import socket
import select
import time
from xml.sax.saxutils import escape
import os

import gtk
import gobject

from kiwi.datatypes import currency
from kiwi.log import Logger
from kiwi.utils import gsignal
from kiwi.ui.entry import KiwiEntry
from kiwi.ui.widgets.label import ProxyLabel

from stoqdrivers.exceptions import DriverError

import stoq
from stoqlib.domain.events import (CardPaymentReceiptPrepareEvent,
                                   CardPaymentReceiptPrintedEvent,
                                   CancelPendingPaymentsEvent,
                                   GerencialReportPrintEvent)
from stoqlib.domain.interfaces import ICompany, ICreditProvider
from stoqlib.domain.person import Person, PersonAdaptToCreditProvider
from stoqlib.domain.payment.method import CreditCardData
from stoqlib.exceptions import DeviceError
from stoqlib.gui.base.dialogs import (BasicDialog, run_dialog,
                                      get_current_toplevel)
from stoqlib.lib.message import yesno, warning
from stoqlib.lib.osutils import get_application_dir
from stoqlib.lib.formatters import get_price_format_str

from teftranslation import tef_gettext as _

log = Logger('stoqlib.plugins.tef.vespague')

# Utility Functions

def wait(secs):
    time_start = time.time()
    time_end = (time_start + secs)

    while time_end > time.time():
        while gtk.events_pending():
            gtk.main_iteration()


# Main classes


class _TemporaryTEFData(object):
    tipos_cartao = {'Credito': CreditCardData.TYPE_CREDIT,
                    'Debito': CreditCardData.TYPE_DEBIT,
                    'Estabelecimento':
                        CreditCardData.TYPE_CREDIT_INSTALLMENTS_STORE,
                    'Administradora':
                        CreditCardData.TYPE_CREDIT_INSTALLMENTS_PROVIDER,
                    # TODO: add TYPE_DEBIT_INSTALLMENTS ?
                    'Parcelado': CreditCardData.TYPE_DEBIT,
                    'Pr\xc3\xa9-datado': CreditCardData.TYPE_DEBIT_PRE_DATED,}

    tipo_cartao = None
    tipo_pagamento = None
    tipo_financiado = None
    parcelas = None
    parcelas_valor = None
    parcelas_vencimento = None
    valor = None
    admnistradora = None
    taxa_servico = None
    rede = None
    vencimento = None
    autorizacao = None
    nsu = None
    comprovante1via = None
    comprovante2via = None

    def get_card_type(self):
        # FIXME: o simulador não retorna valores para testarmos qualquer valor
        # diferente de 'A vista'.

        # This will match TYPE_DEBIT or TYPE_CREDIT
        card_type = self.tipos_cartao.get(self.tipo_cartao)

        if self.tipo_cartao == 'Credito' and self.tipo_pagamento == 'Parcelado':
            card_type = self.tipos_cartao.get(self.tipo_financiado)
        if self.tipo_cartao == 'Debito' and self.tipo_pagamento != 'A vista':
            card_type = self.tipos_cartao.get(self.tipo_pagamento)

        return card_type


class VespagueBackground(gobject.GObject):
    app = stoq.program_name
    version = stoq.version

    default_host = 'localhost'
    default_port = 60906

    gsignal('display-message', str, bool)
    gsignal('hide-message')
    gsignal('ask-value', str, str, str, object, retval=str)
    gsignal('confirm-payment', object, retval=bool)

    sequence = 1
    EOM='\r\n\t\t\r\n\t\t\t\r\n\t\t\r\n\t'

    def __init__(self, host=None, port=None):
        gobject.GObject.__init__(self)
        self.sequence = int(time.time())
        self._cancel_interation = False
        self._can_cancel_interation = False
        self.host = host or self.default_host
        self.port = port or self.default_port
        self.socket = None
        self._pending_flag = os.path.join(get_application_dir(),
                                          'tef_pending_flag')


    #
    # Private
    #

    def _connect(self):
        if self.socket is None:
            try:
                self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.socket.connect((self.host, self.port))
            except Exception, e:
                self.socket = None
                self.display_message(e, show_ok=True)
                return False
        return True

    def _get_new_sequence(self):
        seq = self.sequence
        self.sequence += 1
        return seq

    def _parse_reply(self, reply):
        # Convert the reply into a dict.
        new_reply = {}
        last_key = None
        for line in reply.split('\n'):
            parts = line.split('=')
            if len(parts) == 2:
                if last_key:
                    content = '='.join(parts)
                    new_reply[last_key] += content.replace('"', '')
                    # end of the content of the current key, reset last_key.
                    if content.endswith('"'):
                        last_key = None
                    else:
                        # multiline content, append a newline.
                        new_reply[last_key] += '\n'
                else:
                    new_reply[parts[0]] = parts[1].replace('"', '')
                    # it is a multiline value, set last_key.
                    if not parts[-1].endswith('"'):
                        last_key = parts[0]
                        new_reply[last_key] += '\n'
            elif last_key:
                content = '='.join(parts)
                new_reply[last_key] += content.replace('"', '')
                # same as above.
                if content.endswith('"'):
                    last_key = None
                else:
                    # multiline content, append a newline.
                    new_reply[last_key] += '\n'
        return new_reply

    def _send_message(self, **kwargs):
        msg = ""
        for key, value in kwargs.items():
            msg += '%s="%s"\n' % (key, value)

        log.debug('TEF message sent: %s' % kwargs)

        if not self._connect():
            # eg, send an empty reply if the connection do not succeed
            return {}
        self.socket.send(msg)
        data = ''
        while True:
            read, __, __ = select.select([self.socket], [], [], 0.1)
            if read:
                data += self.socket.recv(1024)
                if not data.endswith(self.EOM):
                    continue
                break
            else:
                while gtk.events_pending():
                    gtk.main_iteration()

                if self._cancel_interation:
                    # cancel_interation once per request.
                    self._cancel_interation = False
                    return {}

        parsed = self._parse_reply(data)
        log.debug('TEF message reply: %s' % parsed)
        return parsed

    def _handle_errors(self, message):
        value = message.get('automacao_coleta_retorno')
        if not value:
            value = message.get('retorno')

        if value == '5':
            user_msg = message.get('automacao_coleta_mensagem')
            if not user_msg:
                user_msg = message.get('mensagem')

            if user_msg:
                # translators: Try no imperativo: 'Tente'
                user_msg += '\n' + _('Try again.')
                self.display_message(user_msg, show_ok=True)

            self.finish(self._get_new_sequence())
            return True

        elif value == '9':
            user_msg = message.get('automacao_coleta_mensagem')
            if self.can_cancel_interation():
                self.set_can_cancel_interation(False)
            if user_msg:
                self.display_message(user_msg, show_ok=True)

            self.finish(self._get_new_sequence())
            return True

        return False

    def _transaction_execute(self, sequence, return_value, transaction,
                             **kwargs):
        # This will start a transaction execution based on the transaction
        # name.
        reply = self._send_message(sequencial=sequence,
                                   retorno=return_value,
                                   servico="executar",
                                   transacao=transaction, **kwargs)
        if not self._handle_errors(reply):
            return reply

    def _transaction_cancel(self, sequence):
        # Cancel the current transaction.
        reply = self._send_message(sequencial=sequence,
                                   automacao_coleta_sequencial=sequence,
                                   automacao_coleta_retorno="9")
        if not self._handle_errors(reply):
            self.finish(self._get_new_sequence())

    def _transaction_loop(self, reply, default_values={}, suggested_values={}):
        # A transaction loop works showing, asking the user for values needed
        # to complete the transaction and handling possible errors.
        # A transaction loop stops when the first expected reply was received
        # from the service.
        # Returns a tuple with the last reply received from the service and
        # the last message sent, if the transaction was well succeed. Returns
        # a tuple with False otherwise.
        while True:
            aut_seq = reply.get('automacao_coleta_sequencial')
            tipo = reply.get('automacao_coleta_tipo')
            msg = reply.get('automacao_coleta_mensagem')
            if not aut_seq:
                break

             # There is no variable to ask the user, only a message to show.
            if msg and not tipo:
                # heuristica para inválido/inválido
                show_ok = bool(re.search(r'inv.+lid.+', msg.lower()))
                self.display_message(msg, show_ok)
                if not show_ok:
                    wait(2)

            value = default_values.get(msg)
            suggested_value = suggested_values.get(msg)
            # Even we have a default value, we should ask at least once for
            # the user.
            if tipo and not value:
                options = reply.get('automacao_coleta_opcao')
                if options:
                    options = options.split(';')
                mask = reply.get('automacao_coleta_mascara', '')
                if suggested_value:
                    msg += '\n' + _('(Default value: %s)') % suggested_value
                value = self.emit('ask-value', msg, tipo, mask, options)
                # The user cancelled the transaction
                if value == 'False':
                    self._transaction_cancel(aut_seq)
                    return False, False

                if not value and suggested_value:
                    value = suggested_value

            params = dict(automacao_coleta_retorno=0,
                          automacao_coleta_sequencial=aut_seq)
            if tipo and value:
                params['automacao_coleta_informacao'] = value

            reply = self._send_message(**params)
            if self._handle_errors(reply):
                return False, False

            return reply, params

        return False, False

    def _confirm_sale_card_payment(self, details):
        tef_data = _TemporaryTEFData()
        try:
            tef_data.tipo_cartao = details['transacao_tipo_cartao']
            tef_data.tipo_pagamento = details['transacao_pagamento']
            # optional, use .get to not raise KeyError
            tef_data.tipo_financiado = details.get('transacao_financiado')
            tef_data.parcelas = details.get('transacao_parcela')
            tef_data.parcelas_valor = details.get('transacao_parcela_valor')
            tef_data.parcelas_vencimento = details.get(
                'transacao_parcela_vencimento')

            tef_data.valor = details.get('transacao_valor')
            tef_data.admnistradora = details.get('transacao_administradora')
            tef_data.taxa_servico = details.get('transacao_valor_taxa_servico')
            tef_data.rede = details.get('transacao_rede')
            tef_data.vencimento = details['transacao_vencimento']
            tef_data.autorizacao = int(details['transacao_autorizacao'])
            tef_data.nsu = int(details['transacao_nsu'])
            tef_data.comprovante1via = details['transacao_comprovante_1via']
            tef_data.comprovante2via = details['transacao_comprovante_2via']
        except KeyError:
            raise
            return

        return self.emit('confirm-payment', tef_data)

    def _get_receipt_from_reply(self, reply):
        if isinstance(reply, bool):
            return ''
        receipt = ''
        if reply.has_key('transacao_comprovante_1via'):
            receipt += reply['transacao_comprovante_1via'] + '\n'
        if reply.has_key('transacao_comprovante_2via'):
            receipt += reply['transacao_comprovante_2via']
        return receipt

    def _print_receipt(self, receipt):
        close_previous = False
        while True:
            try:
                GerencialReportPrintEvent.emit(receipt, close_previous)
                break
            except (DriverError, DeviceError):
                # "Erro na impressão. Deseja tentar novamente?",
                if not yesno(_("Error printing. Would you like to try again?"),
                             # translators: Try no infinitivo: 'Tentar'
                             gtk.RESPONSE_YES, _("Try again"),
                                               _("Cancel")):
                    break

                close_previous = True
                continue

    def _create_pending_tag(self):
        if not os.access(self._pending_flag, os.F_OK):
            log.debug('Creating pending transactions flag')
            os.mknod(self._pending_flag)

    def _remove_pending_tag(self):
        if os.access(self._pending_flag, os.F_OK):
            log.debug('Removing pending transactions flag')
            os.unlink(self._pending_flag)

    def _try_typed_card_sale(self, reply, sequence):
        # Enter in a typed sale, eg where the user types the card number.
        if not reply and self.can_cancel_interation():
            # Not sure why, but the vespague client seems to not handle
            # the following message properly if we don't wait a certain
            # amount of time.
            wait(1)
            return self._send_message(
                automacao_coleta_sequencial=int(sequence) + 1,
                automacao_coleta_retorno=9)
        return reply

    #
    # Public
    #

    def start(self):
        seq = self._get_new_sequence()
        reply = self._send_message(sequencial=seq,
                                   retorno=1,
                                   servico='iniciar',
                                   versao=self.version,
                                   aplicacao=self.app)
        return int(reply.get('retorno') or -1) == 1

    def cancel_interation(self):
        """Requests the current interation to stop. Only one step of the
        interation will be stopped.
        """
        self._cancel_interation = self.can_cancel_interation()

    def can_cancel_interation(self):
        """Returns True if the current interation could be stopped, False
        otherwise.
        """
        return self._can_cancel_interation

    def set_can_cancel_interation(self, can_cancel):
        """Enables the current interation to be stopped.
        """
        self._can_cancel_interation = can_cancel

    def finish(self, sequence):
        reply = {}
        while True:
            reply = self._send_message(sequencial=sequence,
                                       servico='finalizar',
                                       retorno=1)
            if len(reply) > 1:
                break

        return int(reply.get('retorno') or -1) == 1

    def vender_cartao(self, value, boarding_fee="0.00", service_fee="0.00",
                      withdraw="0.00"):
        default_values = {
            'Valor': value,
        }

        suggested_values = {
            'Valor da Entrada': '0.00',
            '1\xc2\xaa Parcela': datetime.date.today().strftime('%d/%m/%Y'),
            'Taxa de Embarque': boarding_fee,
            'Taxa de Servi\xc3\xa7o': service_fee,
            'Valor do Saque': withdraw,
        }

        seq = self._get_new_sequence()
        reply = self._transaction_execute(sequence=seq, return_value=1,
                                          transaction="Cartao Vender")
        if not reply:
            return

        while True:
            msg = reply.get('automacao_coleta_mensagem')
            aut_seq = reply.get("automacao_coleta_sequencial")

            if msg and msg == 'INSIRA OU PASSE O CART\xc3\x83O':
                self.set_can_cancel_interation(True)
            elif msg and msg == 'Parcelas':
                self.set_can_cancel_interation(True)
            else:
                self.set_can_cancel_interation(False)

            if msg and msg == 'C\xc3\xb3digo de Seguran\xc3\xa7a':
                #(I - Ilegível / N - Não possui)
                reply['automacao_coleta_mensagem'] += '\n' + \
                    _('(I - Unreadable / N - Not available)')

            _check1 = 'transação aprovada'
            _check2 = 'transação aceita'
            _check3 = 'transacao aprovada'
            if msg and (_check1 in msg.lower()
                        or _check2 in msg.lower()
                        or _check3 in msg.lower()):
                # Exibindo mensagem APROVADA por 3 segundos
                wait(3)
                break

            reply, sent = self._transaction_loop(reply, default_values,
                                                 suggested_values)

            # It will automatically enter in a typed sale, if needed.
            reply = self._try_typed_card_sale(reply, aut_seq)

            if not reply:
                break

        if not reply:
            return
        # Save last reply details, as it contains nmu, receipt, and every
        # details about the transaction.
        reply, sent = self._transaction_loop(reply)
        NSU_SEQUENCE_NUMBER[int(reply['transacao_nsu'])] = seq
        # Try to confirm the payments (it was already confirmed in vespague).
        if self._confirm_sale_card_payment(reply) is False:
            seq = reply.get('automacao_coleta_sequencial')
            self._transaction_cancel(seq)
            return False

        self._create_pending_tag()
        return True

    def vender_cartao_confirma(self, nsu):
        # finish sale card transaction
        seq = NSU_SEQUENCE_NUMBER.pop(nsu)
        reply = self._transaction_execute(sequence=seq, return_value=0,
                                          transaction="Cartao Vender")
        if not len(NSU_SEQUENCE_NUMBER):
            self._remove_pending_tag()

        # FIXME: george, is this really necessary: ????
        # If there is still one last message, display it.
        msg = reply.get('automacao_coleta_mensagem')
        if msg:
            self.display_message(msg, show_ok=True)
            reply = self._send_message(automacao_coleta_retorno=0,
                                       automacao_coleta_sequencial=1)
        self.hide_message()
        return True

    def vender_cartao_cancela(self, nsu):
        seq = NSU_SEQUENCE_NUMBER.pop(nsu)
        self._transaction_execute(sequence=seq, return_value=9,
                                  transaction="Cartao Vender")
        if not len(NSU_SEQUENCE_NUMBER):
            self._remove_pending_tag()

    def cancelar_venda(self):
        seq = self._get_new_sequence()
        reply = self._transaction_execute(seq, return_value=1,
                                          transaction="Administracao Cancelar")
        if not reply:
            return

        cancel_data = {'Data Transa\xc3\xa7\xc3\xa3o Original': None,
                       'N\xc3\xbamero do Documento': None,
                       'Valor da Transa\xc3\xa7\xc3\xa3o': None}

        last_message = None
        sent = None
        while True:
            #FIXME: find a better way to stop iteration here.
            if reply and 'terminal' in reply.keys():
                break


            aut_seq = reply.get("automacao_coleta_sequencial")
            last_message = reply.get('automacao_coleta_mensagem') or ''

            if last_message and last_message == 'INSIRA OU PASSE O CART\xc3\x83O':
                self.set_can_cancel_interation(True)

            if (last_message.startswith('Solicitando')  and
                all(cancel_data.values())):
                values = []
                for key, value in cancel_data.items():
                    # FIXME: This sale value is not always showing up.
                    if key == 'Valor da Transa\xc3\xa7\xc3\xa3o':
                        value = '%0.2f' % float(value)
                    values.extend([key, value])

                # translators: keep 'Confirmar Cancelamento?'
                msg = '%s: %s\n%s: %s\n%s: %s\n\n'  % tuple(values) + \
                         _('Confirm Cancel?')

                # Translators: Cancelar Venda / Não Cancelar Venda
                if not yesno(msg, gtk.RESPONSE_NO, _('Cancel Sale'),
                                                   _("Don't Cancel Sale")):
                    sequencial = reply.get('automacao_coleta_sequencial')
                    self._transaction_cancel(sequencial)
                    self.hide_message()
                    return

            reply, sent = self._transaction_loop(reply)

            if last_message in cancel_data.keys() and sent:
                cancel_data[last_message] = \
                                    sent.get('automacao_coleta_informacao')
            # XXX: special case, this message appears only in credit card sales
            # and the value should be used instead of the transaction value.
            if last_message == 'Valor Cancelamento':
                cancel_data['Valor da Transa\xc3\xa7\xc3\xa3o'] =  \
                                    sent.get('automacao_coleta_informacao')

            # It will automatically enter in a typed sale, if needed.
            reply = self._try_typed_card_sale(reply, aut_seq)

            if not reply:
                break

        if reply:
            self._transaction_execute(return_value=0,
                                      transaction="Administracao Cancelar",
                                      sequence=seq)

        self.finish(seq)
        comprovante = self._get_receipt_from_reply(reply)
        if comprovante:
            self._print_receipt(comprovante)

    def reimprimir_venda(self):
        seq = self._get_new_sequence()
        reply = self._transaction_execute(seq, return_value=1,
                                          transaction="Administracao Reimprimir")
        if not reply:
            return

        while True:
            #FIXME: find a better way to stop iteration here.
            if reply and 'terminal' in reply.keys():
                break

            reply, sent = self._transaction_loop(reply)
            if not reply:
                break

        if not reply:
            # the transaction was already finished.
            return

        self.finish(seq+1)
        comprovante = self._get_receipt_from_reply(reply)
        if comprovante:
            self._print_receipt(comprovante)

    def transacoes_pendentes(self):
        seq = self._get_new_sequence()
        reply = self._transaction_execute(seq, return_value=1,
                                          transaction="Administracao Pendente",
                                          # This actually disables it
                                          transacao_opcao="Confirmar")
        retval = True

        _check = 'Nao ha transacoes pendentes'
        msg = reply.get('mensagem') or ''
        if msg.startswith(_check):
            self._remove_pending_tag()

        if not reply:
            return

        while True:
            video_msg = reply.get('automacao_coleta_video_mensagem')
            msg = reply.get('automacao_coleta_mensagem')
            if msg and video_msg:
                reply['automacao_coleta_mensagem'] += '\n'.join([video_msg])
            elif video_msg:
                reply['automacao_coleta_mensagem'] = video_msg

            if reply.get('mensagem'):
                self.display_message(reply.get('mensagem'))
                wait(2)
                self.hide_message()
                retval = False

            _check = 'Todas as transações foram resolvidas'
            if msg and _check in msg:
                self._remove_pending_tag()

            reply, sent = self._transaction_loop(reply)
            if not reply:
                break

        self.finish(seq+1)
        return retval

    def hide_message(self):
        self.emit('hide-message')

    def display_message(self, msg, show_ok=False):
        self.emit('display-message', msg, show_ok)


class VespagueDialog(BasicDialog):
    title = 'TEF Stoq'
    size = (600, 300)
    field_size_table = {
        'Digite o n\xc3\xbamero do cart\xc3\xa3o': 19,
        'C\xc3\xb3digo de Seguran\xc3\xa7a': 5,
        'Quatro ultimos digitos': 4,
    }

    def __init__(self, msg, type, mask, options):
        BasicDialog.__init__(self)
        self._initialize(hide_footer=True, size=VespagueDialog.size,
                         title=VespagueDialog.title)
        toplevel = self.get_toplevel()
        toplevel.set_deletable(False)
        self.dtype = type
        self.mask = mask
        long_msg = len(msg.split('\n')) > 3

        if options:
            # When we have only a few options (4), use buttons show the
            # alternatives. Otherwise, let's use a combo box.
            if len(options) <= 4 and not long_msg:
                button_box = gtk.VButtonBox()
                button_box.set_layout(gtk.BUTTONBOX_SPREAD)
                if msg:
                    msg += '\n' + _('(Press ESC to cancel)')
                    button_box.pack_start(self._create_label(msg))

                for opt in options:
                    button = gtk.Button(opt)
                    button.connect('clicked', self._on_button__clicked)
                    button_box.pack_start(button)

                button_box.show_all()
                self.add(button_box)
            else:
                vbox = gtk.VBox(spacing=6)
                if long_msg:
                    table = gtk.Table(1, 2)
                    layout = gtk.Layout(None, None)
                    # unfortunately, (-1, -1) it is not working in this case.
                    layout.set_size(500, 700)
                    table.attach(layout, 0, 1, 0, 1, gtk.FILL | gtk.EXPAND,
                                 gtk.FILL | gtk.EXPAND, 0, 0)
                    vscroll = gtk.VScrollbar(None)
                    table.attach(vscroll, 1, 2, 0, 1, gtk.FILL | gtk.SHRINK,
                                 gtk.FILL | gtk.SHRINK, 0, 0)
                    vadjust = layout.get_vadjustment()
                    vscroll.set_adjustment(vadjust)
                    label = self._create_label(msg)
                    layout.put(label, 150, 0)
                    vbox.pack_start(table, expand=True)
                elif msg:
                    vbox.pack_start(self._create_label(msg))

                hbox = gtk.HBox(spacing=6)
                combo = gtk.combo_box_new_text()
                combo.connect('changed', self._on_combo__changed)
                for opt in options:
                    combo.append_text(opt)
                combo.set_active(0)
                hbox.pack_start(combo)

                ok_button = gtk.Button(stock='gtk-ok')
                ok_button.connect('clicked', self._on_combo_button__clicked)
                hbox.pack_start(ok_button, expand=False)
                vbox.pack_start(hbox, expand=False)
                if not long_msg:
                    vbox.pack_start(self._create_label(' '))
                vbox.show_all()
                self.add(vbox)

        if not options:
            vbox = gtk.VBox(spacing=6)
            hbox = gtk.HBox(spacing=6)
            hbox.set_homogeneous(False)
            # usability
            if type == 'D' and mask:
                msg += '\n(%s)' % mask.lower()

            label = self._create_label(msg)
            hbox.pack_start(label)

            entry = KiwiEntry()

            if type and mask:
                # we will not update self.mask, since the converted mask will
                # be used only in the widget.
                mask = self._convert_mask(type, mask)

            if mask:
                entry.set_mask(mask)
                entry.set_max_length(len(mask))

            if 'senha' in msg.lower():
                entry.set_visibility(False)

            for key in self.field_size_table.keys():
                if msg  == key or msg.startswith(key):
                    entry.set_max_length(self.field_size_table[key])

            entry.connect('activate', self._on_entry__activate)
            hbox.pack_start(entry)

            msg = _('(Press ENTER to confirm)') + '\n' + \
                  _('(Press ESC to cancel)')
            label = self._create_label(msg)
            hbox.pack_start(label)
            vbox.pack_start(hbox)
            vbox.show_all()
            self.add(vbox)

    #
    # Private
    #

    def _convert_mask(self, type, mask):
        new_mask = ''
        if type == 'N':
            new_mask = mask.replace('#', '0')
        elif type == 'X':
            new_mask = mask.replace('X', 'A')
        elif type == 'D':
            new_mask = mask.replace('d', '0')
            new_mask = new_mask.replace('M', '0')
            new_mask = new_mask.replace('y', '0')

        # there is no way to specify a variable number in kiwi masks.
        if new_mask.startswith('.') or new_mask.endswith('.'):
            return ''
        return new_mask

    def _create_label(self, msg):
        label = ProxyLabel()
        label.set_bold(True)
        label.set_text(escape(msg))
        return label

    def _validate_date(self, date):
        # Checks if the date is valid or not.
        # Returns the same date string if the date is valid or an empty string
        # otherwise.
        mask = self.mask
        # converts the mask to the format required by datetime.
        mask = mask.replace('yyyy', '%Y')
        mask = mask.replace('yy', '%y')
        mask = mask.replace('MM', '%m')
        mask = mask.replace('dd', '%d')
        try:
            datetime.datetime.strptime(date, mask)
        except ValueError:
            return ""
        return date

    #
    # BasicDialog
    #

    def confirm(self, *args):
        self.close()

    def cancel(self, *args):
        # set retval to detect the ESC usage.
        self.retval = False
        return False

    #
    # Callbacks
    #

    def _on_button__clicked(self, widget):
        self.retval = widget.get_label()
        self.confirm()

    def _on_combo_button__clicked(self, widget):
        self.confirm()

    def _on_combo__changed(self, widget):
        self.retval = widget.get_active_text()

    def _on_entry__activate(self, widget):
        if self.dtype == 'D':
            self.retval = self._validate_date(widget.get_text())
        else:
            self.retval = widget.get_text()
        self.confirm()


# Global holding the receipts and each payment created.
NSU_TEF_DATA = {}
NSU_PAYMENTS = {}
NSU_SEQUENCE_NUMBER ={}
# The maximum quantity of cards allowed in a sale
MAX_MULTIPLE_CARDS = 3


class Vespague(object):

    def __init__(self):
        self.background = VespagueBackground()
        self.background.connect('display-message',
                                self._on_background__display_message)
        self.background.connect('hide-message',
                                self._on_background__hide_message)
        self.background.connect('ask-value', self._on_background__ask_value)
        self.background.connect('confirm-payment',
                                self._on_background__confirm_payment)

        CardPaymentReceiptPrepareEvent.connect(
            self._on_CardPaymentReceiptPrepareEvent)
        CardPaymentReceiptPrintedEvent.connect(
            self._on_CardPaymentReceiptPrintedEvent)
        CancelPendingPaymentsEvent.connect(self._on_CancelPendingPaymentsEvent)

        self._is_showing_dialog = False
        self.message_dialog = None

    def start(self):
        self._is_ok_button_visible = False
        return self.background.start()

    def finish(self):
        seq = self.background._get_new_sequence()
        return self.background.finish(seq)

    #
    # Public
    #

    def run_sale_card_dialog(self, sale, method, value=None):
        if not self._can_perform_sale_card(sale.payments):
            self._show_message(
                _('Card limit reached: %d') % MAX_MULTIPLE_CARDS,
                  show_ok=True)
            return

        self.start()
        self._sale = sale
        self._method = sale.get_connection().get(method)
        if not value:
            value = sale.get_total_sale_amount()

        total = get_price_format_str() % value
        self._show_waiting_message()
        details = self.background.vender_cartao(total)
        self._hide_message()
        return details

    def run_cancel_sale_dialog(self):
        self.start()
        self._show_waiting_message()
        retval = self.background.cancelar_venda()
        self._hide_message()
        return retval

    def run_reprint_sale_dialog(self):
        self.start()
        self._show_waiting_message()
        retval = self.background.reimprimir_venda()
        self._hide_message()
        return retval

    def run_pending_transactions_dialog(self):
        self.start()
        self._show_waiting_message()
        retval = self.background.transacoes_pendentes()
        self._hide_message()
        return retval

    #
    # Private
    #

    def _create_sale_payment(self, tef_data):
        provider = self._get_or_create_provider(tef_data.rede)
        due_date = dateutil.parser.parse(tef_data.vencimento, dayfirst=True)
        value = currency(tef_data.valor)

        log.debug('Created TEF Payment: %s, %s, %s' %
                                    (self._sale.group, value, due_date))
        inpayment = self._method.create_inpayment(self._sale.group, value,
                                                  due_date)

        tipo_cartao = tef_data.get_card_type()
        operation = self._method.operation
        payment = inpayment.get_adapted()
        data = operation.get_card_data_by_payment(payment)
        data.card_type = tipo_cartao
        data.provider = provider
        if tef_data.parcelas:
            data.installments = int(tef_data.parcelas)
        else:
            data.installments = 1

        if tef_data.parcelas_valor:
            data.entrance_value = currency(tef_data.parcelas_valor)

        try:
            data.fee = provider.get_fee_for_payment(provider, data)
        except Exception:
            data.fee = 0

        data.fee_value = data.fee * payment.value / 100
        data.nsu = tef_data.nsu
        data.auth = tef_data.autorizacao

        # Save payment for later confimation/cancelation, after the
        # receipts are printed
        NSU_PAYMENTS.setdefault(data.nsu, []).append(payment)

        # Save the tef_data so we can print the correct receipt later
        NSU_TEF_DATA[tef_data.nsu] = tef_data

        return bool(inpayment)

    def _get_or_create_provider(self, provider_name):
        conn = self._sale.get_connection()
        provider = PersonAdaptToCreditProvider.selectOneBy(
            short_name=provider_name.upper(),
            connection=conn)
        if not provider:
            person = Person(name=provider_name, connection=conn)
            person.addFacet(ICompany, connection=conn)
            provider = person.addFacet(ICreditProvider,
                                    short_name=provider_name.upper(),
                                    open_contract_date=datetime.date.today(),
                                    connection=conn)
            log.debug('Creating credit provider:' % provider_name.upper())

        return provider

    def _can_perform_sale_card(self, sale_payments):
        return len(NSU_PAYMENTS.keys()) < MAX_MULTIPLE_CARDS

    def _delete_event(self, widget, event, data=None):
        # Disable cancel when showing a message (using ESC key), the user
        # should cancel the operation using the pinpad.
        if self.background.can_cancel_interation():
            self.background.cancel_interation()
        return True

    def _create_message_dialog(self):
        self.message_dialog = gtk.Dialog(title='TEF Stoq',
                                         parent=get_current_toplevel(),
                                         flags=gtk.DIALOG_MODAL)
        self.message_dialog.connect('delete-event', self._delete_event)
        self.message_ok_button = self.message_dialog.add_button(
            gtk.STOCK_OK, gtk.RESPONSE_CLOSE)
        self.message_ok_button.connect('clicked',
                            self._on_message_ok_button__clicked)
        self.message_dialog.set_deletable(False)
        self.message_dialog.set_position(gtk.WIN_POS_CENTER_ON_PARENT)
        self.message_dialog.set_size_request(600, 300)

        self.message_label = ProxyLabel('')
        self.message_label.set_bold(True)
        self.message_dialog.vbox.pack_start(self.message_label)

    def _show_waiting_message(self):
        self._show_message(_('Waiting TEF...'))

    def _show_message(self, message, show_ok=False):
        if not self.message_dialog:
            self._create_message_dialog()
        self._is_ok_button_visible = show_ok
        if self.background.can_cancel_interation():
            message += '\n' + _('(Press ESC to cancel)')

        self.message_label.set_text(escape(message))
        self.message_dialog.show_all()

        if show_ok is False:
            self.message_ok_button.hide()
        else:
            self.message_ok_button.grab_focus()

        # wait until the user confirm the message
        while self._is_ok_button_visible:
            wait(0.01)
            time.sleep(0.01)
        return True

    def _hide_message(self):
        self.message_label.set_text('')
        self.message_dialog.hide()
        self._is_ok_button_visible = False

    def _confirm_payment(self, nsu):
        payments = NSU_PAYMENTS.pop(nsu)
        self.background.vender_cartao_confirma(nsu)
        today = datetime.date.today()
        for payment in payments:
            due_date = payment.due_date.date()
            if not payment.is_paid() and due_date == today:
                payment.pay()

    def _cancel_payment(self, nsu):
        payments = NSU_PAYMENTS.pop(nsu)
        self.background.vender_cartao_cancela(nsu)
        for payment in payments:
            payment.cancel()

    #
    # Signals
    #

    def _on_background__display_message(self, background, message, show_ok):
        self._is_showing_dialog = True
        self._show_message(message, show_ok)

    def _on_background__hide_message(self, background):
        self._is_showing_dialog = False
        self._hide_message()

    def _on_background__ask_value(self, background, msg, type, mask, options):
        if not self._is_ok_button_visible:
            self._hide_message()
        # otherwise we have to wait the user confirm the last message.
        while self._is_ok_button_visible:
            wait(0.01)
            time.sleep(0.01)

        # If there is only one option, there is no reason to show the dialog
        if options and not isinstance(options, list):
            return options
        elif isinstance(options, list) and len(options) == 1:
            return options[0]

        self._is_showing_dialog = True
        retval = run_dialog(VespagueDialog, get_current_toplevel(), msg, type,
                            mask, options)
        if isinstance(retval, str):
            retval = retval.replace(',', '.')
        self._is_showing_dialog = False
        self._show_waiting_message()
        return retval

    def _on_background__confirm_payment(self, background, tef_data):
        return self._create_sale_payment(tef_data)

    def _on_CardPaymentReceiptPrepareEvent(self, nsu, supports_duplicate):
        # should return the receipt
        tef_data = NSU_TEF_DATA.get(nsu)
        if tef_data:
            return ''.join([tef_data.comprovante1via,
                            tef_data.comprovante2via])

    def _on_CardPaymentReceiptPrintedEvent(self, nsu):
        self._confirm_payment(nsu)

    def _on_CancelPendingPaymentsEvent(self):
        pending_nsu = NSU_PAYMENTS.keys()
        if not len(pending_nsu):
            return

        for nsu in pending_nsu:
            self._cancel_payment(nsu)

        if len(pending_nsu) == 1:
            # Ultima transação TEF não foi completada com sucesso.
            msg = _('Last TEF transaction was not sucessfully completed.')
        elif len(pending_nsu) > 1:
            msg = (_('Last %s TEF transactions wore not sucessfully '
                     'completed.') % len(pending_nsu))

        #'Favor, reter o cupom fiscal'
        msg += '\n' + _('Please, retaing fiscal coupon')
        warning(msg)

    def _on_message_ok_button__clicked(self, widget):
        self._hide_message()
