#!/bin/bash

# References:
# https://wiki.edubuntu.org/Plymouth
# https://help.ubuntu.com/community/LiveCDCustomization
# https://wiki.ubuntu.com/CustomizeLiveInitrd

# LiveCD we're using
UBUNTUCD=ubuntu-12.04-desktop-i386.iso
SQUASHFS=casper/filesystem.squashfs
MANIFEST=casper/filesystem.manifest

# FIXME: Installed kernel version instead of hard-coded
KERNELVER=3.2.0-24-generic
KERNELVER=3.2.0-23-generic-pae

# FIXME: We should use lsb_release here
RELEASENAME=precise

# List of packages to add
PKGS_INSTALL=pkgs.install

# List of packages to remove
PKGS_REMOVE=pks.remove

# Cannot contain letters, used by wubi
STOQ_VERSION=1.6.0

LIVEUSER=stoq
DBNAME=$LIVEUSER
TOPDIR=`pwd`

# What to build from bzr
: ${KIWI_REVISION:="tag:latest"}
: ${KIWI_BRANCH:="lp:kiwi"}
: ${STOQDRIVERS_REVISION:="tag:latest"}
: ${STOQDRIVERS_BRANCH:="lp:stoqdrivers"}
: ${STOQ_REVISION:="tag:latest"}
: ${STOQ_BRANCH:="lp:stoq"}
: ${DEBUG_OPTIONS:=""}
: ${PUBLIC_ISO_BASE_URL:="http://download.stoq.com.br/livecd"}

_query() {
    sudo -u postgres psql -Aqt $DBNAME -c "$1"
}

_check_started() {
    if [ ! -e edit/$0 ]; then
        echo 'You must use start first'
        exit 1
    fi

    if [ $0 -nt edit/$0 ]; then
        echo "Updating $0"
        cp $0 edit
    fi
}

_check_ended() {
    if [ -e edit/$0 ]; then
        echo 'You must use end first'
        exit 1
    fi
}

# Deletes previous ppa, packages and enables universe if disabled
_prepare_apt() {
    rm -f edit/etc/apt/sources.list.d/stoq-*
    sed -i "s/ restricted$/ restricted universe/" edit/etc/apt/sources.list
    chroot edit apt-get remove -qq --purge --yes $PKGS_REMOVE || true
}

_apt_update() {
    _check_started
    chroot edit apt-get -q update
}

_apt_upgrade() {
    _check_started
    chroot edit apt-get upgrade --yes
}


# Commands

bootstrap_stable() {
    build_debs
    extract
    start
    remove_packages
    use_stable
    install_stoq
    fixes
    configure_stoq
    enable_ecf
    #enable_tef
    install_slideshow
    initrd
    end
    build
}

bootstrap_unstable() {
    build_debs
    extract
    start
    remove_packages
    use_unstable
    install_stoq
    fixes
    configure_stoq
    enable_ecf
    #enable_tef
    install_slideshow
    initrd
    end
    build
}

_build_from_bzr() {
    MODULE=$1
    BRANCH=$2
    REVISION=$3

    cd $TOPDIR/bzr
    if [ -d $MODULE ]; then
       cd $MODULE
       bzr revert
       # Overwrite in case we switch branches in the UI
       bzr pull --overwrite $BRANCH
    else
       bzr branch $BRANCH -r $REVISION $MODULE
       cd $MODULE
    fi
    echo "$MODULE $BRANCH `bzr revno`" >> ../revnos.txt
    debuild --no-tgz-check -uc -us
}

_build_from_source() {
    PACKAGE=$1
    PACKAGE_PATH=$TOPDIR/packages/$1

    if [ ! -d $PACKAGE_PATH ]; then
        echo "ERROR: No package called $PACKAGE found"
        exit 1
    fi

    cd $PACKAGE_PATH
    debuild --no-tgz-check -uc -us -b
    cd $TOPDIR
}

build_debs() {
    rm -f $TOPDIR/packages/plymouth-theme-stoq*.deb     \
          $TOPDIR/packages/ubiquity-slideshow-stoq*.deb\
          $TOPDIR/packages/*.dsc     \
          $TOPDIR/packages/*.build   \
          $TOPDIR/packages/*.changes \
          $TOPDIR/packages/*.tar.gz
    _build_from_source plymouth-theme-stoq
    _build_from_source ubiquity-slideshow-stoq

    mkdir -p $TOPDIR/bzr
    rm -f $TOPDIR/bzr/*.deb \
          $TOPDIR/bzr/*.dsc \
          $TOPDIR/bzr/*.build \
          $TOPDIR/bzr/*.changes \
          $TOPDIR/bzr/*.tar.gz \
          $TOPDIR/bzr/revnos.txt
    _build_from_bzr kiwi $KIWI_BRANCH $KIWI_REVISION
    _build_from_bzr stoqdrivers $STOQDRIVERS_BRANCH $STOQDRIVERS_REVISION
    _build_from_bzr stoq $STOQ_BRANCH $STOQ_REVISION
    rm -fr $TOPDIR/edit/bzr
    mkdir -p $TOPDIR/edit/bzr
    cp $TOPDIR/bzr/*.deb $TOPDIR/edit/bzr
    cd $TOPDIR
}

extract() {
    if [ ! -x edit/bin/bash ]; then
        # Mount the Desktop .iso
        if [ ! -d mnt ]; then
            mkdir mnt
        fi

        # Extract .iso contents into dir 'CD'
        mount -o loop $UBUNTUCD mnt
        if [ ! -d CD ]; then
            mkdir CD
            rsync --exclude=/$SQUASHFS -a mnt/ CD
        fi

        # Extract the SquashFS filesystem
        unsquashfs -n mnt/$SQUASHFS
        mv squashfs-root edit
        umount mnt
    fi
}

start() {
    cp /etc/resolv.conf edit/etc/
    cp /etc/hosts edit/etc/
    [ ! -c edit/dev/null ] && sudo mount --bind /dev/ edit/dev
    [ ! -f edit/var/cache/apt/archives/.is-cache ] && sudo mount --bind cache edit/var/cache/apt/archives
    # Hard link ourselves in the chroot
    [ ! -f edit/$0 ] && ln $0 edit
    chmod 777 edit/root
    mkdir -p edit/root/.stoq
    chmod 777 edit/root/.stoq

    # This is temporary only for stoqdbadmin to run within the
    # chroot, it'll be replaced later on.
    cat > edit/root/.stoq/stoq.conf << EOF
[General]

[Database]
rdbms = postgres
address =
port = 5432
dbname = stoq
dbusername = postgres
EOF
    chmod 777 edit/root/.stoq/stoq.conf
    sudo chroot edit $0 chroot_start
}

remove_packages() {
    _check_started
    PKGS=$(cat pkgs.remove|xargs echo)
    chroot edit apt-get remove -qq --purge --yes $PKGS 2> /dev/null
}

use_stable() {
    _check_started
    _prepare_apt
    chroot edit apt-add-repository -y ppa:stoq-dev/lancamentos
    _apt_update
}

use_unstable() {
    _check_started
    _prepare_apt
    chroot edit apt-add-repository -y ppa:stoq-maintainers/unstable
    _apt_update
}

chroot_fixes() {
    export HOME=/root

    # Install postgresql
    # Will install:
    #   libpq5 postgresql-8.4 postgresql-client-8.4 postgresql-client-common
    #   postgresql-common

    if [ ! -e /etc/init.d/postgresql ]; then
        apt-get install --yes postgresql
    fi

    # Install any deb that was copied to /
    total=$(ls -1 /*deb | wc -l)
    if [ "$total" -gt "0" ]; then
        for x in $(ls /*deb || echo ''); do
            dpkg -i $x
            rm $x
        done
    fi

    # Enable stoq auto start
    auto_dir=/etc/skel/.config/autostart
    auto_file="$auto_dir/stoq.desktop"
    mkdir -p $auto_dir
    if [ ! -e $auto_file ]; then
    cat > $auto_file <<EOF
[Desktop Entry]
Type=Application
Name=Stoq
Exec=/usr/bin/stoq
Icon=/usr/share/stoq/pixmaps/stoq.png
Comment=
X-GNOME-Autostart-enabled=true
EOF
    fi

    # Add stoq to the desktop
    desktop_dir=/etc/skel/Desktop
    desktop_file="$desktop_dir/stoq.desktop"
    mkdir -p $desktop_dir
    cp /usr/share/applications/stoq.desktop $desktop_dir
}

enable_ecf() {
    _check_started
    chroot edit $0 chroot_enable_ecf
}

enable_tef() {
    _check_started
    cp packages/stoqtef*.deb edit/
    chroot edit $0 chroot_enable_tef
}

install_slideshow() {
    _check_started
    cp packages/ubiquity-slideshow-stoq*.deb edit/

    # We cannot use apt-get to remove this as it will have dependency problems
    chroot edit bash -c "dpkg -r ubiquity-slideshow-ubuntu"
    chroot edit bash -c "dpkg -i ubiquity-slideshow-stoq*.deb"
    chroot edit bash -c "rm -rf ubiquity-slideshow-stoq*.deb"
}

install_stoq() {
    _check_started
    PKGS=$(cat pkgs.install|xargs echo)
    chroot edit apt-get install -qq --yes $PKGS || true

    [ -d edit/bzr ] && chroot edit bash -c "dpkg -i /bzr/*.deb" || true
}


fixes() {
    _check_started
    # First start boot script
    cp -r data/stoq-bootstrap edit/etc/init.d
    ln -sf /etc/init.d/stoq-bootstrap edit/etc/rc2.d/S99stoq-bootstrap
    chroot edit $0 chroot_fixes
}

configure_stoq() {
    _check_started
    chroot edit $0 chroot_configure_stoq `hostname`
    cp -r data/home/ubuntu/.stoq/ edit/etc/skel
}

build() {
    _check_ended

    build_squashfs
    build_iso
}

build_squashfs() {
    # Generate list of packages
    chmod +w CD/$MANIFEST
    chroot edit dpkg-query -W --showformat='${Package} ${Version}\n' > CD/$MANIFEST
    cp CD/$MANIFEST CD/$MANIFEST-desktop

    # packages to remove after instalation
    sed -i '/ubiquity/d' CD/$MANIFEST-desktop
    sed -i '/casper/d' CD/$MANIFEST-desktop
    sed -i '/stoqtef/d' CD/$MANIFEST-desktop

    # Create LiveCD filesystem
    [ -f CD/$SQUASHFS ] && rm CD/$SQUASHFS
    SQUASHFS_OPTIONS="-no-progress"
    if [[ "$DEBUG_OPTIONS" == *"fastsquashfs"* ]]; then
        SQUASHFS_OPTIONS="$SQUASHFS_OPTIONS -noD -noF"
    fi
    mksquashfs edit/ CD/$SQUASHFS $SQUASHFS_OPTIONS

    printf $(sudo du -sx --block-size=1 edit | cut -f1) > CD/casper/filesystem.size
}

generate_metadata() {
    ISO=$1

    # create a link and set permissions
    cd iso
    ln -sf $ISO stoq-LATEST-livecd.iso

    # md5sum
    md5sum $ISO > $ISO.md5sum

    # wubi metadata
    cat > $ISO.metalink << EOF
<?xml version="1.0" encoding="utf-8"?>
<?xml-stylesheet type="text/xsl" href="http://download.eclipse.org/eclipse/metalink.xsl"?>

<metalink version="3.0" pubdate="$(date)" xmlns="http://www.metalinker.org/">
  <publisher>
    <name>Stoq</name>
    <url>http://www.stoq.com.br/</url>
  </publisher>
  <version>$STOQ_VERSION</version>
  <description>Stoq</description>
  <logo>http://www.stoq.com.br/images/stoq.png</logo>
  <files>
    <file name="$ISO">
      <size>$(stat -c "%s" $ISO)</size>
      <os>Linux-x86</os>
      <verification>
        <hash type="md5">$(cut -d\  -f1 $ISO.md5sum)</hash>
      </verification>
      <resources maxconnections="1">
        <url type="http" location="br" preference="100">$PUBLIC_ISO_BASE_URL/$ISO</url>
      </resources>
    </file>
  </files>
</metalink>
EOF

    # bzr revisions
    cp ../bzr/revnos.txt $ISO.manifest
    chmod 644 $ISO
    cd ..
}

build_iso() {
    #VERSION=`chroot edit dpkg-query -W  --showformat='${Version}\n' stoq`
    VERSION=$STOQ_VERSION
    DEST_NAME=stoq-${VERSION}-livecd.iso

    # Stoq customizations
    cp isolinux/* CD/isolinux

    # md5sum
    [ -f CD/md5sum.txt ] && rm CD/md5sum.txt
    (cd CD && find -type f -print0 | xargs -0 md5sum | grep -v \
                isolinux/boot.cat > md5sum.txt)

    # This must match wubi's regexp, which is like this:
    # (?P<name>[\w\s-]+) (?P<version>[\d.]+)(?: LTS)?
    # \"(?P<codename>[\D]+)\" - (?P<subversion>[\D]+) (?P<arch>i386|amd64)
    # \((?P<build>[\d.]+)\)
    echo -n "Stoq $STOQ_VERSION \"Stoq\" - Release i386 ($(date +%Y%m%d))" > CD/.disk/info

    # Create LiveCD Image
    cd CD && \
    mkisofs -r -V "STOQ_LIVECD" -quiet -cache-inodes -J -l -b \
        isolinux/isolinux.bin -c isolinux/boot.cat -no-emul-boot \
        -boot-load-size 4 -boot-info-table -o "../iso/$DEST_NAME" .

    cd ..
    generate_metadata $DEST_NAME
}

initrd() {
    [ ! -f edit/$0 ] && ln $0 edit
    cp packages/plymouth*deb edit/
    sudo chroot edit $0 chroot_initrd
    mv edit/initrd.lz CD/casper/
}

chroot_initrd() {
    [ ! -f /proc/uptime ] && mount -t proc none /proc

    sed -i 's/maybe-ubiquity/DO-NOT-RUN-maybe-uniquity/' /etc/init/ubiquity.conf

    sed -i 's/USERNAME="ubuntu"/USERNAME="stoq"/' /etc/casper.conf
    sed -i 's/"Live/"Stoq Live/' /etc/casper.conf
    sed -i 's/\# export FLAVOUR="Ubuntu"/export FLAVOUR="Stoq"/' /etc/casper.conf

    dpkg -i plymouth*.deb
    rm -fr plymount*.deb
    update-alternatives --set default.plymouth /lib/plymouth/themes/stoq/stoq.plymouth
    update-alternatives --set text.plymouth /lib/plymouth/themes/stoq/stoq-text.plymouth

    mkinitramfs -o /initrd.lz -c lzma $KERNELVER
}

end() {
    _check_started;
    sudo chroot edit $0 chroot_end
    rm edit/$0
    [ -c edit/dev/pts/ptmx ] && umount edit/dev
    sudo umount edit/var/cache/apt/archives || true
}


test_plymouth() {
    plymouth --show-splash ; \
    for ((I=0; I<10; I++)); do \
        plymouth --update=test$I ; \
        sleep 1; \
    done; \
    #plymouth quit
}

# Chroot

chroot_start() {
    [ ! -f /proc/uptime ] && mount -t proc none /proc
    [ ! -d /sys/kernel ] && mount -t sysfs none /sys
    [ ! -c /dev/pts/ptmx ] && mount -t devpts none /dev/pts
    export HOME=/root
    dbus-uuidgen > /var/lib/dbus/machine-id
    dpkg-divert --local --rename --add /sbin/initctl
    [ ! -f /sbin/initctl ] && ln -s /bin/true /sbin/initctl
    [ -e /etc/init.d/postgresql ] && /etc/init.d/postgresql start
}

chroot_end() {
    # http://stackoverflow.com/a/6852484/14337
    _query "UPDATE pg_class SET relowner = (SELECT oid FROM pg_roles WHERE rolname = '$LIVEUSER')
 WHERE relname IN (SELECT relname FROM pg_class, pg_namespace
                    WHERE pg_namespace.oid = pg_class.relnamespace AND
                          pg_namespace.nspname = 'public');"
    [ -e /etc/init.d/postgresql ] && /etc/init.d/postgresql stop

    [ -e /etc/hosts ] && rm /etc/hosts
    [ -e /etc/resolv.conf ] && rm -f /etc/resolv.conf
    [ -e /var/lib/dbus/machine-id ] && rm -f /var/lib/dbus/machine-id
    [ -L /sbin/initctl ] && rm -f /sbin/initctl
    dpkg-divert --rename --remove /sbin/initctl
    umount /proc || umount -lf /proc || true
    umount /sys || true
    umount /dev/pts || true
    rm -rf /tmp/* ~/.bash_history
}


chroot_configure_stoq() {
    HOSTNAME=$1
    export HOME=/root

    [ -e /etc/init.d/postgresql ] && /etc/init.d/postgresql start

    sudo -u postgres stoqcreatedbuser $LIVEUSER
    sudo -u postgres stoqdbadmin init -e --force

    # update machine name
    _query "UPDATE branch_station SET name='ubuntu' WHERE name='$HOSTNAME'"

    # update admin user password
    _query "UPDATE login_user SET pw_hash=md5('admin') WHERE username='admin'"

    # set demo mode
    _query "UPDATE parameter_data SET field_value='1' WHERE field_name='DEMO_MODE'"

    # Enable online services
    _query "UPDATE parameter_data SET field_value='1' WHERE \
            field_name='ONLINE_SERVICES'"

    rm -fr /etc/skel/.stoq
}

chroot_enable_ecf() {
    export HOME=/root

    res=$(_query "select 1 from installed_plugin where plugin_name='ecf'";)
    if [[ $res != 1 ]]; then
        echo 'Enabling plugin ecf'

        sudo -u postgres stoqdbadmin enable_plugin ecf
        _query \
        "INSERT INTO ecf_printer (te_id, model, brand,
            device_name, device_serial, station_id, is_active, baudrate) VALUES (2,
            'Simple', 'virtual', '', 'Virtual', 2, 't', 115200);"

        printer_id=$(_query "SELECT id FROM ecf_printer WHERE model='Simple';")

        cat <<EOF | sudo -u postgres psql stoq
SELECT pg_catalog.setval('device_constant_id_seq', 20, true);
INSERT INTO device_constant VALUES (8, 2, 1, 'Substitution', NULL, 41, 'TS', $printer_id);
INSERT INTO device_constant VALUES (9, 3, 1, 'Exemption', NULL, 42, 'TE', $printer_id);
INSERT INTO device_constant VALUES (10, 4, 1, 'No tax', NULL, 43, 'TN', $printer_id);
INSERT INTO device_constant VALUES (11, 5, 1, '18 %', 18.00, 45, 'T1', $printer_id);
INSERT INTO device_constant VALUES (12, 6, 1, '12 %', 12.00, 45, 'T2', $printer_id);
INSERT INTO device_constant VALUES (13, 7, 1, '5 %', 5.00, 45, 'T3', $printer_id);
INSERT INTO device_constant VALUES (14, 8, 1, '3 %', 3.00, 44, 'S0', $printer_id);
INSERT INTO device_constant VALUES (15, 9, 2, 'Cheque', 3.00, 1, 'C', $printer_id);
INSERT INTO device_constant VALUES (16, 10, 2, 'Boleto', 3.00, 2, 'B', $printer_id);
INSERT INTO device_constant VALUES (17, 11, 2, 'Cartão Débito', 3.00, 4, 'D', $printer_id);
INSERT INTO device_constant VALUES (18, 12, 2, 'Financeira', 3.00, 5, 'F', $printer_id);
INSERT INTO device_constant VALUES (19, 13, 2, 'Dinheiro', 3.00, 0, 'M', $printer_id);
INSERT INTO device_constant VALUES (20, 14, 2, 'Cartão Crédito', 3.00, 3, 'R', $printer_id);
EOF

    fi
}

chroot_enable_tef() {
    export HOME=/root
    export LC_ALL=pt_BR.UTF8
    export LANG=pt_BR.UTF8
    export LANGUAGE=pt_BR.UTF8

    if [ ! -e /usr/bin/java ]; then
        echo 'Java not found. Installing'
        apt-get install librxtx-java openjdk-6-jre --no-install-recommends \
            --yes
    fi

    if [ -e stoqtef_*deb ]; then
        dpkg -i stoqtef_*.deb
        rm stoqtef_*.deb
    fi

    res=$(_query "select 1 from installed_plugin where plugin_name='tef'";)
    if [[ $res != 1 ]]; then
        echo 'Enabling plugin tef'
        sudo -u postgres stoqdbadmin enable_plugin tef
        _query "INSERT INTO tef_client (te_created_id, te_modified_id, \
            path, host, port) VALUES (1, 2, '', 'localhost', 60906) "
    fi
}

function usage() {
    echo $"Usage: $0 [command]"
    echo "List of commands:"
    echo
    declare -F|grep -v chroot_|egrep -v "^declare -f _"|cut -d\  -f3-|sed "s/^/    /"
    echo
}

if [ "$(declare -F | egrep ^declare\ -f\ $1\$)" = "" ]; then
    echo "ERROR: no such command: $1"
    usage
    exit 1
fi

unset LANG
unset LANGUAGE
unset LC_ALL
unset LC_MESSAGES
unset LC_NUMERIC

set -x -e

$@

exit 0
